from typing import Dict, List, Literal, Union

from dataclasses import Field
from datetime import datetime
from decimal import Decimal
from enum import IntEnum

from fastapi_utils.api_model import APIModel
from thanakan.models.bankcode import AnyBankCode, BankCode
from thanakan.services.model.common import ProxyType, SlipData


class StatusCode(IntEnum):
    """
    only known status codes
    """

    success = 1000


class Status(APIModel):
    code: Union[StatusCode, int]
    description: str


class CredentialsData(APIModel):
    access_token: str
    token_type: Literal["Bearer"] = "Bearer"
    expires_in: int
    expires_at: int


class BaseResponse(APIModel):
    status: Status
    data: Dict


class SCBCredentialsResponse(BaseResponse):
    data: CredentialsData


class QR30Data(APIModel):
    qr_raw_data: str
    qr_image: str  # seems like a base64 string


class CreateQR30Response(BaseResponse):
    data: QR30Data


class VerifyResponse(BaseResponse):
    data: SlipData


class TransactionInquiryResponse(BaseResponse):
    data: List[Dict]


class WebhookBody(APIModel):
    transaction_date_and_time: datetime = Field(alias='transactionDateandTime')
    sending_bank_code: Union[BankCode, AnyBankCode]
    payer_name: str
    payer_proxy_type: ProxyType
    payer_proxy_id: str
    payer_account_number: str
    receiving_bank_code: Union[BankCode, AnyBankCode]
    payee_name: str
    payee_proxy_type: ProxyType
    payee_proxy_id: str
    payee_account_number: str
    amount: Decimal
    currency_code: str  # wtf "764", thb?
    transaction_id: str
    transaction_type: str
    bill_payment_ref1: str
    bill_payment_ref2: str
    bill_payment_ref3: str
    channel_code: str

    class Config:
        schema_extra = {
            "example": {
                "transactionDateandTime": "2021-02-24T11:10:12+07:00",
                "sendingBankCode": "014",
                "payerName": "Susophida Khunnawut",
                "payerProxyType": "ACCOUNT",
                "payerProxyId": "1502830001",
                "payerAccountNumber": "1502830001",
                "receivingBankCode": "014",
                "payeeName": "TestBiller1608347078",
                "payeeProxyType": "BILLERID",
                "payeeProxyId": "925244165291982",
                "payeeAccountNumber": "0987654321",
                "amount": "100",
                "currencyCode": "764",
                "transactionId": "9ed5caf2-0a38-4f5a-80b1-20e76f26b2a6",
                "transactionType": "Domestic Transfers",
                "billPaymentRef1": "ABCDEFGHIJ1234567890",
                "billPaymentRef2": "ABCDEFGHIJ1234567890",
                "billPaymentRef3": "YQH",
                "channelCode": "PMH",
            }
        }



from typing import Literal, Optional

import uuid
from decimal import Decimal

import httpx
from furl import furl
from httpx._types import CertTypes
from pydantic import constr, validate_arguments
from thanakan.services.base import BankApi
from thanakan.services.model.scb import (
    BaseResponse,
    CreateQR30Response,
    SCBCredentialsResponse,
    StatusCode,
    TransactionInquiryResponse,
    VerifyResponse,
)

SANDBOX_BASE_URL = "https://api-sandbox.partners.scb/partners/sandbox/"


class SCBAPI(BankApi):
    creds: Optional[SCBCredentialsResponse] = None

    def __init__(
        self,
        application_key,
        application_secret,
        cert: Optional[CertTypes] = None,
        base_url=SANDBOX_BASE_URL,
    ):
        self.application_key = application_key
        self.application_secret = application_secret
        self.cert = cert
        self.base_url = furl(base_url)
        auth_url = self.base_url / "oauth/token"
        client = httpx.Client(cert=self.cert)
        # auth = OAuth2ClientCredentials(
        #     auth_url.url,
        #     client_id=self.consumer_id,
        #     client_secret=self.consumer_secret,
        #     client=client,
        # )
        self.client = httpx.AsyncClient(
            base_url=base_url,
            cert=self.cert,
            # auth=auth
        )

        self.client_sync = httpx.Client(
            base_url=base_url,
            cert=self.cert,
            # auth=auth
        )

    async def get_token(self):
        """ get access_token from scb """
        # TODO: turn this into a  custom auth engine
        body = {
            "applicationKey": self.application_key,
            "applicationSecret": self.application_secret,
        }
        headers = {
            "Content-Type": "application/json",
            "resourceOwnerId": self.application_key,
            "requestUId": uuid.uuid4().hex,
            "accept-language": "EN",
        }

        auth_url = self.base_url / "v1/oauth/token"
        r = httpx.post(
            auth_url.url,
            json=body,
            headers=headers,
            # auth=(self.consumer_id, self.consumer_secret),
            cert=self.cert,
        )

        if r.status_code == 200:
            self.creds = SCBCredentialsResponse.parse_raw(r.content)
            return self.creds
        else:
            raise ConnectionError(r.json())

    @validate_arguments
    async def create_qr30(
        self,
        pp_id: constr(min_length=15, max_length=15),
        amount: Decimal,
        ref1: constr(max_length=20, regex="[A-Z0-9]*"),
        ref2: constr(max_length=20, regex="[A-Z0-9]*"),
        ref3: constr(max_length=20, regex="[A-Z]{3}[A-Z0-9]*"),
        qr_type: Literal["PP"] = "PP",
        pp_type: Literal["BILLERID"] = "BILLERID",
    ):
        headers = {
            "Content-Type": "application/json",
            "authorization": f"Bearer {self.creds.data.access_token}",
            "resourceOwnerId": self.application_key,
            "requestUId": uuid.uuid4().hex,
            "accept-language": "EN",
        }
        payload = {
            "qrType": qr_type,
            "ppType": pp_type,
            "amount": str(amount),
            "ppId": pp_id,
            "ref1": ref1,
            "ref2": ref2,
            "ref3": ref3,
        }
        r = await self.client.post(
            "/v1/payment/qrcode/create", json=payload, headers=headers
        )

        if r.status_code == 200:
            parsed_response = CreateQR30Response.parse_raw(r.content)
            if parsed_response.status.code == StatusCode.success:
                return parsed_response
            else:
                raise ConnectionError(parsed_response.status.description)
        else:
            raise ConnectionError(r.json())

    @validate_arguments
    async def verify_slip(
        self,
        transaction_ref_id: str,
        sending_bank_id: str,
    ):
        headers = {
            "authorization": f"Bearer {self.creds.data.access_token}",
            "resourceOwnerId": self.application_key,
            "requestUId": uuid.uuid4().hex,
            "accept-language": "EN",
        }

        the_furl = (
            furl("/v1/payment/billpayment/transactions") / transaction_ref_id
        )
        the_furl.args["sendingBank"] = sending_bank_id

        r = await self.client.get(the_furl.url, headers=headers)

        if r.status_code == 200:
            parsed_response = VerifyResponse.parse_raw(r.content)
            if parsed_response.status.code == StatusCode.success:
                return parsed_response
            else:
                raise ConnectionError(parsed_response.status.description)
        else:
            raise ConnectionError(r.json())

    @validate_arguments
    async def query_transaction(
        self,
        biller_id: str,
        reference1: str,
        reference2: str,
        transaction_date: constr(regex=r"\d{4}-\d{2}-\d{2}"),
        amount: Decimal,
        event_code: Literal["00300100", "00300104"] = "00300100",
    ):

        the_furl = furl("/v1/payment/billpayment/inquiry")
        the_furl.add(
            args={
                "eventCode": event_code,
                "billerId": biller_id,
                "reference1": reference1,
                "reference2": reference2,
                "amount": f"{amount:.2f}",
                "transactionDate": transaction_date,
            }
        )

        headers = {
            "authorization": f"Bearer {self.creds.data.access_token}",
            "resourceOwnerId": self.application_key,
            "requestUId": uuid.uuid4().hex,
            "accept-language": "EN",
        }

        r = await self.client.get(the_furl.url, headers=headers)

        if r.status_code == 200:
            # TODO: parse into specific object
            parsed_response = TransactionInquiryResponse.parse_raw(r.content)
            if parsed_response.status.code == StatusCode.success:
                return parsed_response
            else:
                raise ConnectionError(parsed_response.status.description)
        else:
            raise ConnectionError(r.json())

    @validate_arguments
    async def create_deeplink(
        self,
    ):
        import requests

        url = "https://api-sandbox.partners.scb/partners/sandbox/v3/deeplink/transactions"

        payload = '{\n\t"transactionType": "PURCHASE",\n\t"transactionSubType": ["BP", "CCFA", "CCIPP"],\n\t"sessionValidityPeriod": 60,\n\t"sessionValidUntil": "",\n\t"billPayment": {\n\t\t"paymentAmount": 100,\n\t\t"accountTo": "123456789012345",\n\t\t"accountFrom": "123451234567890",\n\t\t"ref1": "ABCDEFGHIJ1234567890",\n\t\t"ref2": "ABCDEFGHIJ1234567890",\n\t\t"ref3": "ABCDEFGHIJ1234567890"\n\t},\n\t"creditCardFullAmount": {\n\t\t"merchantId": "1234567890ABCDEF",\n\t\t"terminalId": "1234ABCD",\n\t\t"orderReference": "12345678",\n\t\t"paymentAmount": 100\n\t},\n\t"installmentPaymentPlan": {\n\t\t"merchantId": "4218170000000160",\n\t\t"terminalId": "56200004",\n\t\t"orderReference": "AA100001",\n\t\t"paymentAmount": 10000.00,\n\t\t"tenor": "12",\n\t\t"ippType": "3",\n\t\t"prodCode": "1001"\n\t},\n\t"merchantMetaData": {\n\t\t"callbackUrl": "",\n\t\t"merchantInfo": {\n\t\t\t"name": "SANDBOX MERCHANT NAME"\n\t\t},\n\t\t"extraData": {},\n\t\t"paymentInfo": [\n\t\t\t{\n\t\t\t\t"type": "TEXT_WITH_IMAGE",\n\t\t\t\t"title": "",\n\t\t\t\t"header": "",\n\t\t\t\t"description": "",\n\t\t\t\t"imageUrl": ""\n\t\t\t},\n\t\t\t{\n\t\t\t\t"type": "TEXT",\n\t\t\t\t"title": "",\n\t\t\t\t"header": "",\n\t\t\t\t"description": ""\n\t\t\t}\n\t\t]\n\t}\n}'
        payload2 = {
            "transactionType": "PURCHASE",
            "transactionSubType": ["BP", "CCFA", "CCIPP"],
            "sessionValidityPeriod": 60,
            "sessionValidUntil": "",
            "billPayment": {
                "paymentAmount": 100,
                "accountTo": "123456789012345",
                "accountFrom": "123451234567890",
                "ref1": "ABCDEFGHIJ1234567890",
                "ref2": "ABCDEFGHIJ1234567890",
                "ref3": "ABCDEFGHIJ1234567890",
            },
            "creditCardFullAmount": {
                "merchantId": "1234567890ABCDEF",
                "terminalId": "1234ABCD",
                "orderReference": "12345678",
                "paymentAmount": 100,
            },
            "installmentPaymentPlan": {
                "merchantId": "4218170000000160",
                "terminalId": "56200004",
                "orderReference": "AA100001",
                "paymentAmount": 10000.00,
                "tenor": "12",
                "ippType": "3",
                "prodCode": "1001",
            },
            "merchantMetaData": {
                "callbackUrl": "",
                "merchantInfo": {"name": "SANDBOX MERCHANT NAME"},
                "extraData": {},
                "paymentInfo": [
                    {
                        "type": "TEXT_WITH_IMAGE",
                        "title": "",
                        "header": "",
                        "description": "",
                        "imageUrl": "",
                    },
                    {
                        "type": "TEXT",
                        "title": "",
                        "header": "",
                        "description": "",
                    },
                ],
            },
        }
        headers = {
            "Content-Type": "application/json",
            "authorization": "Bearer <Your Access Token>",
            "resourceOwnerId": "<Your API Key>",
            "requestUId": "fe2f0e21-361b-44fd-85e8-75734184eb34",
            "channel": "scbeasy",
            "accept-language": "EN",
            "Cookie": "TS01e7ba6b=01e76b033c782fc88aa5838c33fda6eb0cf563d307a7b3cb6f17b84b610aecd9e7ef0d4e0fef15be922efd3884bc2a1fa1eb1404da",
        }

        response = requests.request("POST", url, headers=headers, data=payload)

        print(response.text)

    @validate_arguments
    async def get_deeplink(self, transaction_ref_id: str):
        import requests

        url = "https://api-sandbox.partners.scb/partners/sandbox/v2/transactions/9ed5caf2-0a38-4f5a-80b1-20e76f26b2a6"

        payload = {}
        headers = {
            "authorization": "Bearer ",
            "resourceOwnerId": "",
            "requestUId": "d03b9b3b-a275-4f78-a7c4-1475a9e65ff6",
            "accept-language": "EN",
        }

        response = requests.request("GET", url, headers=headers, data=payload)

        print(response.text)



"""

ref.
https://apiportal.kasikornbank.com/product/public/Information/Slip%20Verification/Try%20API/d519307a-6d82-4e77-b9f4-dc74e542c742
"""
from typing import Dict, Optional

import uuid
from datetime import datetime

import httpx
import pytz
from furl import furl
from httpx._types import CertTypes
from httpx_auth import OAuth2ClientCredentials
from loguru import logger
from thanakan.models.bankcode import AnyBankCode
from thanakan.services.base import BankApi
from thanakan.services.model.kbank import VerifyResponse

bkk_tz = pytz.timezone("Asia/Bangkok")


class KBankAPI(BankApi):
    creds: Optional[Dict] = None

    def __init__(
        self,
        consumer_id,
        consumer_secret,
        cert: CertTypes,
        base_url="https://openapi.kasikornbank.com",
    ):
        self.consumer_id = consumer_id
        self.consumer_secret = consumer_secret
        self.cert = cert
        self.base_url = furl(base_url)
        auth_url = self.base_url / "oauth/token"
        client = httpx.Client(cert=self.cert)
        # change this to async with https://docs.authlib.org/en/latest/client/httpx.html
        auth = OAuth2ClientCredentials(
            auth_url.url,
            client_id=self.consumer_id,
            client_secret=self.consumer_secret,
            client=client,
        )
        self.client = httpx.AsyncClient(
            base_url=base_url, cert=self.cert, auth=auth
        )

        self.client_sync = httpx.Client(
            base_url=base_url, cert=self.cert, auth=auth
        )

    async def get_token(self):

        body = {"grant_type": "client_credentials"}

        auth_url = self.base_url / "oauth/token"
        r = httpx.post(
            auth_url.url,
            data=body,
            auth=(self.consumer_id, self.consumer_secret),
            cert=self.cert,
        )

        if r.status_code == 200:
            self.creds = r.json()
            return r.json()
        else:
            return r

    async def verify_slip(self, sending_bank_id, trans_ref, *, raw=True):
        body = {
            "rqUID": uuid.uuid4().hex,
            "rqDt": datetime.now(tz=bkk_tz).isoformat(),
            "data": {"sendingBank": sending_bank_id, "transRef": trans_ref},
        }

        r = await self.client.post("/v1/verslip/kbank/verify", json=body)

        if r.status_code == 200:
            json = r.json()
            if raw:
                return json
            try:
                return VerifyResponse(**json)
            except Exception as e:
                logger.debug("data is {}", json)
                raise e
        else:
            return r

    def verify_slip_sync(self, sending_bank_id, trans_ref, *, raw=True):
        body = {
            "rqUID": uuid.uuid4().hex,
            "rqDt": datetime.now(tz=bkk_tz).isoformat(),
            "data": {"sendingBank": sending_bank_id, "transRef": trans_ref},
        }

        r = self.client_sync.post("/v1/verslip/kbank/verify", json=body)

        if r.status_code == 200:
            json = r.json()
            if raw:
                return json
            try:
                return VerifyResponse(**json)
            except Exception as e:
                logger.debug("data is {}", json)
                raise e
        else:
            return r
